import os
import typing

import chromadb
from chromadb import API
from chromadb.api.models.Collection import Collection
from chromadb.utils.embedding_functions import SentenceTransformerEmbeddingFunction
from pydantic_settings import BaseSettings
from tqdm import tqdm
from loguru import logger

from srctag.model import FileContext, RuntimeContext


class StorageConfig(BaseSettings):
    db_path: str = "./chroma"
    collection_name: str = "default_collection"

    # English: paraphrase-MiniLM-L6-v2
    # Multi langs: paraphrase-multilingual-MiniLM-L12-v2
    st_model_name: str = "paraphrase-MiniLM-L6-v2"


class Storage(object):
    def __init__(self, config: StorageConfig = None):
        if not config:
            config = StorageConfig()
        self.config = config

        self.chromadb: typing.Optional[API] = None
        self.chromadb_collection: typing.Optional[Collection] = None

    def init_chroma(self):
        if self.chromadb and self.chromadb_collection:
            return
        self.chromadb = chromadb.PersistentClient(path=self.config.db_path)
        self.chromadb_collection = self.chromadb.get_or_create_collection(
            self.config.collection_name,
            embedding_function=SentenceTransformerEmbeddingFunction(
                model_name=self.config.st_model_name
            ),
        )

    def embed_file(self, file: FileContext):
        self.init_chroma()
        sentences = [each.message.split(os.linesep)[0] for each in file.commits]

        self.chromadb_collection.add(
            documents=[os.linesep.join(sentences)],
            metadatas=[{"source": file.name}],
            ids=[file.name],
        )

    def embed_ctx(self, ctx: RuntimeContext):
        self.init_chroma()
        logger.debug("start embedding source files")
        for each_file in tqdm(ctx.files.values()):
            self.embed_file(each_file)
        logger.debug("embedding finished")



import csv
import typing

from chromadb import QueryResult, Metadata
from pydantic import BaseModel
from pydantic_settings import BaseSettings

from srctag.storage import Storage

# tag name -> distance score
SingleTagResult = typing.Dict[str, float]


class TagResult(BaseModel):
    files: typing.Dict[str, SingleTagResult] = dict()

    def export_csv(self, path: str = "srctag-output.csv") -> None:
        file_list = self.files.keys()
        col_list = set().union(*[d.keys() for d in self.files.values()])
        with open(path, "w", newline="", encoding="utf-8-sig") as file:
            writer = csv.writer(file)

            header = [""] + list(col_list)
            writer.writerow(header)

            for each_file in file_list:
                row = [each_file] + [
                    self.files[each_file].get(subkey, "-1") for subkey in col_list
                ]
                writer.writerow(row)


class TaggerConfig(BaseSettings):
    tags: typing.Set[str] = set()

    # eg: 0.3 == choosing the closest 30% results
    n_percent: float = 0.3


class Tagger(object):
    """load feature map, search related files, and tag them"""

    def __init__(self, config: TaggerConfig = None):
        if not config:
            config = TaggerConfig()
        self.config = config

    def tag(self, storage: Storage) -> TagResult:
        storage.init_chroma()
        file_count = storage.chromadb_collection.count()
        n_results = int(file_count * self.config.n_percent)

        ret = dict()
        for each_tag in self.config.tags:
            query_result: QueryResult = storage.chromadb_collection.query(
                query_texts=each_tag,
                n_results=n_results,
                include=["metadatas", "distances"],
            )
            files: typing.List[Metadata] = query_result["metadatas"][0]
            distances: typing.List[float] = query_result["distances"][0]

            minimum = min(distances)
            maximum = max(distances)
            if maximum == minimum:
                # all values are the same
                normalized_scores = [1 for _ in distances]
            else:
                normalized_scores = [
                    1 - ((x - minimum) / (maximum - minimum)) for x in distances
                ]

            for each_file, each_score in zip(files, normalized_scores):
                each_file_name = each_file["source"]
                if each_file_name not in ret:
                    ret[each_file_name] = dict()
                ret[each_file_name][each_tag] = each_score

        return TagResult(files=ret)
